/*
 * plan13.c
 *
 * This is a C port of G3RUH's PLAN-13, a satellite position calculation
 * program.
 *
 * Edson Pereira, PY2SDR
 *
 * History:
 *
 * 08 May 1997 - Finished port of plan13 from BASIC code
 * 16 May 1997 - Added NASA 2-Line file (tle.dat) handling
 * 15 May 2013 - Updated cmd line arguments
 *
 * Program expects TLEs in the file nasabare.txt
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <strings.h>
#include <stdint.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>

#define TRUE   1
#define ONEPPM 1.0e-6
#define LIGHTSPEED 299792.438

struct tm *Tm;
struct timespec ts;

double rxFrequency;
double txFrequency;
double rx, tx;
double observer_lon;
double observer_lat;

char tlefile[] = "nasabare.txt";

double   YM;			       /* Days in a year                     */
double   EL;			       /* Elevation                          */
double   TN;			       /*                                    */

double   E;
double   N;
double   AZ;
double   SLON;
double   SLAT;
double   RR;

double   CL;
double   CS;
double   SL;
double   CO;
double   SO;
double   RE;
double   FL;
double   RP;
double   XX;
double   ZZ;
double   D;
double   R;
double   Rx;
double   Ry;
double   Rz;
double   Ex;
double   Ey;
double   Ez;
double   Ny;
double   Nx;
double   Nz;
double   Ox;
double   Oy;
double   Oz;
double   U;
double   Ux;
double   Uy;
double   Uz;
double   YT;
double   WW;
double   WE;
double   W0;
double   VOx;
double   VOy;
double   VOz;
double   DE;
double   GM;
double   J2;
double   N0;
double   A0;
double   b0;
double   SI;
double   CI;
double   PC;
double   QD;
double   WD;
double   DC;
double   YG;
double   G0;
double   MAS0;
double   MASD;
double   INS;
double   CNS;
double   SNS;
double   EQC1;
double   EQC2;
double   TEG;
double   GHAE;
double   MRSE;
double   MASE;
double   ax;
double   ay;
double   az;
int      OLDRN;

double   T;
double   DT;
double   KD;
double   KDP;
double   M;
int      DR;
long     RN;
double   EA;
double   C;
double   S;
double   DNOM;
double   A;
double   B;
double   RS;
double   Sx;
double   Sy;
double   Sz;
double   Vx;
double   Vy;
double   Vz;
double   AP;
double   CW;
double   SW;
double   RAAN;
double   CQ;
double   SQ;
double   CXx;
double   CXy;
double   CXz;
double   CYx;
double   CYy;
double   CYz;
double   CZx;
double   CZy;
double   CZz;
double   SATx;
double   SATy;
double   SATz;
double   ANTx;
double   ANTy;
double   ANTz;
double   VELx;
double   VELy;
double   VELz;
double   Ax;
double   Ay;
double   Az;
double   Sz;
double   Vz;
double   GHAA;

double   DS;
double   DF;

/* keplerians */

char     SAT[20];
long     SATNO;
double   YE;
double   TE;
double   IN;
double   RA;
double   EC;
double   WP;
double   MA;
double   MM;
double   M2;
long     RV;
double   ALON;
double   ALAT;
double   FR;

/* location */
char     LOC[20];
double   LA;
double   LO;
double   HT;

double      HR;			       /* Hours */
double      DN;

/*
 * Convert degrees to radians
 */
double rad(double deg)
{
    return (M_PI / 180.0 * deg);
}

double deg(double rad)
{
    return (rad * 180.0 / M_PI);
}

double FNatn(double y, double x)
{
    float a;

    if (x != 0)
    {
        a = atan(y / x);
    }
    else
    {
        a = M_PI / 2.0 * sin(y);
    }

    if (x < 0)
    {
        a = a + M_PI;
    }

    if (a < 0)
    {
        a = a + 2.0 * M_PI;
    }

    return (a);
}

/* Convert date to day number
 *
 * Function returns a general day number from year, month, and day.
 * Value is (JulianDay - 1721409.5) or (AmsatDay + 722100)
 *
 */
double FNday(int year, int month, int day)
{
    double JulianDate;

    if (month <= 2)
    {
        year -= 1;
        month += 12;
    }

    JulianDate = (int)(year * YM) + (int)((month + 1) * 30.6) + (day - 428);

    return (JulianDate);
}

double getElement(char *gstr, int gstart, int gstop)
{
    double retval;
    int    k, glength;
    char   gestr[80];

    glength = gstop - gstart + 1;

    for (k = 0; k <= glength; k++)
    {
        gestr[k] = gstr[gstart+k-1];
    }

    gestr[glength] = '\0';
    retval = atof(gestr);
    return(retval);
}

void readElements(char *satellite)
{
    FILE *tle;
    char buffer[81];
    char line1[81];
    char line2[81];

    if ((tle = fopen(tlefile, "r")) == NULL)
    {
        perror( tlefile );
        exit(0);
    }

    for(;;)
    {
        if (fgets(buffer, 80, tle) == NULL)
        {
            printf("Satellite not found...\n");
            exit(0);
        }

        if (buffer[0] == '1' || buffer[0] == '2') continue;
        if (strncasecmp(buffer, satellite, strlen(satellite)) == 0)
        {
            printf("%s", buffer);
            fgets(line1, 80, tle);
            fgets(line2, 80, tle);
            printf("%s%s\n", line1, line2);

            /*
            YE = 1997.0;
            TE = 126.71404377;
            IN = 98.5440;
            RA = 201.9624;
            EC = 0.0009615;
            WP = 356.5498;
            MA = 3.5611;
            MM = 14.27977735;
            M2 = 2.500E-07;
            RV = 18818;
            ALON = 180.0;
            ALAT = 0.0;
            */

            SATNO = (long) (getElement(line1, 3, 7) + ONEPPM);
            RV    = (long) (getElement(line2,64,68) + ONEPPM);
            YE    = getElement(line1,19,20) + 2000;
            TE    = getElement(line1,21,32);
            IN    = getElement(line2, 9,16);
            RA    = getElement(line2,18,25);
            EC    = getElement(line2,27,33) * 1.0e-7;
            WP    = getElement(line2,35,42);
            MA    = getElement(line2,44,51);
            MM    = getElement(line2,53,63);
            M2    = getElement(line1,34,43);

	    /*
            printf("RV %ld\n", RV);
            printf("YE %d\n", (int)YE);
            printf("TE %f\n", TE);
            printf("IN %f\n", IN);
            printf("RA %f\n", RA);
            printf("EC %f\n", EC);
            printf("WP %f\n", WP);
            printf("MA %f\n", MA);
            printf("MM %f\n", MM);
            printf("M2 %f\n\n", M2);
	    */
	    
            break;
        }
    }
}

void init(void)
{
    readElements(SAT);

    /* Observer's location */
    LA = rad( observer_lat );
    LO = rad( observer_lon );
    HT = 20.0 / 1000.0;

    CL = cos(LA);
    SL = sin(LA);
    CO = cos(LO);
    SO = sin(LO);

    /* WGS-84 Earth Ellipsoid */
    RE = 6378.137;
    FL = 1.0 / 298.257222;

    RP = RE * (1.0 - FL);
    XX = RE * RE;
    ZZ = RP * RP;

    D = sqrt( XX * CL * CL + ZZ * SL * SL );

    Rx = XX / D + HT;
    Rz = ZZ / D + HT;

    /* Observer's unit vectors Up EAST and NORTH in geocentric coordinates */
    Ux = CL * CO;
    Ex = -SO;
    Nx = -SL * CO;

    Uy = CL * SO;
    Ey = CO;
    Ny = -SL * SO;

    Uz = SL;
    Ez = 0;
    Nz = CL;

    /* Observer's XYZ coordinates at earth's surface */
    Ox = Rx * Ux;
    Oy = Rx * Uy;
    Oz = Rz * Uz;

    /* Convert angles to radians, etc. */
    RA = rad(RA);
    IN = rad(IN);
    WP = rad(WP);
    MA = rad(MA);
    MM = MM * 2.0 * M_PI;
    M2 = M2 * 2.0 * M_PI;

    YM = 365.25;		    /* Mean year, days                           */
    YT = 365.2421970;		/* Tropical year, days                       */
    WW = 2.0 * M_PI / YT;   /* Earth's rotation rate, rads/whole day     */
    WE = 2.0 * M_PI + WW;	/* Earth's rotation rate, rads/day           */
    W0 = WE / 86400;		/* Earth's rotation rate, rads/sec           */

    /* Observer's velocity, geocentric coordinates */
    VOx = -Oy * W0;
    VOy = Ox * W0;

    /* Convert satellite epoch to Day No. and fraction of a day */
    DE = FNday(YE, 1, 0) + (int)TE;
    TE = TE - (int)TE;

    /* Average Precession rates */
    GM = 3.986E5;               /* Earth's gravitational constant km^3/s^2   */
    J2 = 1.08263E-3;            /* 2nd Zonal coeff, Earth's gravity Field    */
    N0 = MM / 86400;                    /* Mean motion rads/s                */
    A0 = pow(GM / N0 / N0, 1.0 / 3.0);  /* Semi major axis km                */
    b0 = A0 * sqrt(1.0 - EC * EC);      /* Semi minor axis km                */
    SI = sin(IN);
    CI = cos(IN);
    PC = RE * A0 / (b0 * b0);
    PC = 1.5 * J2 * PC * PC * MM;       /* Precession const, rad/day         */
    QD = -PC * CI;                      /* Node Precession rate, rad/day     */
    WD = PC *(5.0 * CI*CI - 1.0) / 2.0; /* Perigee Precession rate, rad/day  */
    DC = -2.0 * M2 / MM / 3.0;          /* Drag coeff                        */

    /* Sideral and solar data. Never needs changing. Valid to year 2000+ */

    /* GHAA, Year YG, Jan 0.0 */
    YG = 2005;
    G0 = 99.7599;
    /* MA Sun and rate, deg, deg/day */
    MAS0 = 356.7424;
    MASD = 0.98560028;
    /* Sun's inclination */
    INS = rad(23.4386);
    CNS = cos(INS);
    SNS = sin(INS);
    /* Sun's equation of center terms */
    EQC1 = 0.03341;
    EQC2 = 0.00035;

    /* Bring Sun data to satellite epoch */
    TEG = (DE - FNday(YG, 1, 0)) + TE;  /* Elapsed Time: Epoch - YG          */
    GHAE = rad(G0) + TEG * WE;          /* GHA Aries, epoch                  */
    MRSE = rad(G0) + (TEG * WW) + M_PI; /* Mean RA Sun at Sat Epoch          */
    MASE = rad(MAS0 + MASD * TEG);      /* Mean MA Sun                       */

    /* Antenna unit vector in orbit plane coordinates */
    CO = cos(rad(ALON));
    SO = sin(rad(ALON));
    CL = cos(rad(ALAT));
    SL = sin(rad(ALAT));
    ax = -CL * CO;
    ay = -CL * SO;
    az = -SL;

    /* Miscellaneous */
    OLDRN = -99999;
}

/*
 * Calculate satellite position at DN, TN
 */
void satvec(void)
{
    T = (DN - DE) + (TN - TE);          /* Elapsed T since epoch             */

    DT = DC * T / 2.0;                  /* Linear drag terms                 */
    KD = 1.0 + 4.0 * DT;
    KDP = 1.0 - 7.0 * DT;
    M = MA + MM * T * (1.0 - 3.0 * DT); /* Mean anomaly at YR,/ TN           */
    DR = (int)(M / (2.0 * M_PI));       /* Strip out whole no of revs        */
    M = M - DR * 2.0 * M_PI;            /* M now in range 0 - 2PI            */
    RN = RV + DR + 1;                   /* Current orbit number              */

    /* Solve M = EA - EC * sin(EA) for EA given M, by Newton's method        */
    EA = M;                             /* Initail solution                  */
    do
    {
        C = cos(EA);
        S = sin(EA);
        DNOM = 1.0 - EC * C;
        D = (EA - EC * S - M) / DNOM;   /* Change EA to better resolution    */
        EA = EA - D;                    /* by this amount until converged    */
    }
    while (fabs(D) > 1.0E-5);

    /* Distances */
    A = A0 * KD;
    B = b0 * KD;
    RS = A * DNOM;

    /* Calculate satellite position and velocity in plane of ellipse */
    Sx = A * (C - EC);
    Vx = -A * S / DNOM * N0;
    Sy = B * S;
    Vy = B * C / DNOM * N0;

    AP = WP + WD * T * KDP;
    CW = cos(AP);
    SW = sin(AP);
    RAAN = RA + QD * T * KDP;
    CQ = cos(RAAN);
    SQ = sin(RAAN);

    /* Plane -> celestial coordinate transformation, [C] = [RAAN]*[IN]*[AP] */
    CXx = CW * CQ - SW * CI * SQ;
    CXy = -SW * CQ - CW * CI * SQ;
    CXz = SI * SQ;
    CYx = CW * SQ + SW * CI * CQ;
    CYy = -SW * SQ + CW * CI * CQ;
    CYz = -SI * CQ;
    CZx = SW * SI;
    CZy = CW * SI;
    CZz = CI;

    /* Compute satellite's position vector, ANTenna axis unit vector    */
    /*   and velocity  in celestial coordinates. (Note: Sz = 0, Vz = 0) */
    SATx = Sx * CXx + Sy * CXy;
    ANTx = ax * CXx + ay * CXy + az * CXz;
    VELx = Vx * CXx + Vy * CXy;
    SATy = Sx * CYx + Sy * CYy;
    ANTy = ax * CYx + ay * CYy + az * CYz;
    VELy = Vx * CYx + Vy * CYy;
    SATz = Sx * CZx + Sy * CZy;
    ANTz = ax * CZx + ay * CZy + az * CZz;
    VELz = Vx * CZx + Vy * CZy;

    /* Also express SAT, ANT, and VEL in geocentric coordinates */
    GHAA = GHAE + WE * T;       /* GHA Aries at elaprsed time T */
    C = cos(-GHAA);
    S = sin(-GHAA);
    Sx = SATx * C - SATy * S;
    Ax = ANTx * C - ANTy * S;
    Vx = VELx * C - VELy * S;
    Sy = SATx * S + SATy * C;
    Ay = ANTx * S + ANTy * C;
    Vy = VELx * S + VELy * C;
    Sz = SATz;
    Az = ANTz;
    Vz = VELz;
}

/*
 * Compute and manipulate range/velocity/antenna vectors
 */
void rangevec(void)
{
    /* Range vector = sat vector - observer vector */
    Rx = Sx - Ox;
    Ry = Sy - Oy;
    Rz = Sz - Oz;

    R = sqrt(Rx * Rx + Ry * Ry + Rz * Rz);    /* Range Magnitute */

    /* Normalize range vector */
    Rx = Rx / R;
    Ry = Ry / R;
    Rz = Rz / R;
    U = Rx * Ux + Ry * Uy + Rz * Uz;
    E = Rx * Ex + Ry * Ey;
    N = Rx * Nx + Ry * Ny + Rz * Nz;

    AZ = deg(FNatn(E, N));
    EL = deg(asin(U));

    /* Solve antenna vector along unit range vector, -r.a = cos(SQ) */
    /*    SQ = deg(acos(-(Ax * Rx + Ay * Ry + Az * Rz))); */

    /* Calculate sub-satellite Lat/Lon */
    SLON = deg(FNatn(Sy, Sx));   /* Lon, + East  */
    SLAT = deg(asin(Sz / RS));   /* Lat, + North */

    /* Resolve Sat-Obs velocity vector along unit range vector. (VOz = 0) */
    RR = (Vx - VOx) * Rx + (Vy - VOy) * Ry + Vz * Rz; /* Range rate, km/sec */
    FR = rxFrequency * (1 - RR / 299792);
}

void sunvec(void)
{

}

void printdata(void)
{
    printf("AZ:%0.1f EL:%0.1f - RX:%f TX:%f\r", AZ, EL, rx, tx);
    fflush(stdout);
}

int main(int argc, char *argv[])
{
    time_t t;
    double nsec;
    
    unsigned char az, el;

    if (argc < 2)
    {
        printf("usage: track <satellite> [duration]\n");
        exit(0);
    }

    strcpy(SAT, argv[1]);

    rxFrequency = 435.0;
    txFrequency = 146.0;
    observer_lon = -48.3750; /* lon east is positive, west is negative */
    observer_lat = -23.1038;

    init();

    while(1)
    {
        timespec_get(&ts, TIME_UTC);
        Tm = gmtime(&ts.tv_sec);
        nsec = ts.tv_nsec/1000000000.0;
        DN =  FNday(Tm->tm_year + 1900, Tm->tm_mon + 1, Tm->tm_mday);
        TN = (Tm->tm_hour + Tm->tm_min/60.0 + (Tm->tm_sec + nsec)/3600.0) / 24;

        satvec();
        rangevec();
        sunvec();

        rx = rxFrequency * (1.0 - RR / LIGHTSPEED);
        tx = txFrequency * (1.0 + RR / LIGHTSPEED);

        printdata();
        usleep(500000);
    }
}

